//go:generate go run syscalltrace_generate_windows.go -output zsyscalltrace_windows.go os_windows syscalltrace_windows

package runtime

import "unsafe"

//go:cgo_import_dynamic runtime._CreateFileW CreateFileW%7 "kernel32.dll"
//go:cgo_import_dynamic runtime._GetEnvironmentVariableW GetEnvironmentVariableW%3 "kernel32.dll"
//go:cgo_import_dynamic runtime._GetModuleFileNameA GetModuleFileNameA%3 "kernel32.dll"

var (
	_CreateFileW,
	_GetEnvironmentVariableW,
	_GetModuleFileNameA,
	_ stdFunction
)

var (
	syscallTraceFileHandle uintptr
	currentProcessName     [_MAX_PATH]byte
)

// syscallTrace logs a syscall trace.
// The trace is written as a comma-separated list of the process name,
// module name, procedure name, and a newline character.
// If the underlying file supports atomic writes, the trace is guaranteed
// to be written atomically.
// procName should be a pointer to a null-terminated string.
func syscallTrace(modHandle uintptr, procName *byte) {
	if syscallTraceFileHandle == _INVALID_HANDLE_VALUE {
		return
	}
	if syscallTraceFileHandle == 0 {
		// It is safe to call initSyscallTrace without locking
		// because the first time syscallTrace is called is during
		// runtime initialization, when there is only one goroutine.
		if !initSyscallTrace() {
			return
		}
	}
	// We can't allocate memory here, as syscallTrace might be called
	// before the memory allocator is initialized. Instead, we use a
	// fixed-size buffer on the stack. The buffer is large enough to
	// hold two commas, one end of line character, and MAX_PATH
	// characters for each of line entry.
	buf := make([]byte, _MAX_PATH*3+3)
	n := findnull(&currentProcessName[0])
	copy(buf, currentProcessName[:n])
	buf[n] = ','
	n += 1
	// Use GetModuleFileNameA instead of GetModuleFileNameW to avoid
	// having to allocate memory for the UTF-16 string.
	n1 := stdcall3(_GetModuleFileNameA, modHandle, uintptr(unsafe.Pointer(&buf[n])), _MAX_PATH)
	if n1 == 0 || n1 == _MAX_PATH {
		// Failed or short buffer.
		buf[n] = '?'
		n += 1
	} else {
		n += int(n1)
	}
	buf[n] = ','
	n += 1
	procNameLength := findnull(procName)
	if procNameLength > _MAX_PATH {
		// Truncate long procedure names,
		// we can't allocate memory to log the full name.
		procNameLength = _MAX_PATH
	}
	copy(buf[n:], unsafe.Slice(procName, procNameLength))
	n += procNameLength
	buf[n] = '\n'
	n += 1
	write(syscallTraceFileHandle, unsafe.Pointer(&buf[0]), int32(n))
}

// initSyscallTrace initializes syscall tracing.
func initSyscallTrace() bool {
	// Get the name of the current process.
	n := int(stdcall3(_GetModuleFileNameA, 0, uintptr(unsafe.Pointer(&currentProcessName[0])), uintptr(len(currentProcessName))))
	if n == 0 || n == len(currentProcessName) {
		// Failed or short buffer.
		currentProcessName[0] = '?'
	}
	// Check if the MS_GO_SYSCALL_TRACE environment variable is set,
	// and if so, use it as the trace file.
	envName := [...]uint16{'M', 'S', '_', 'G', 'O', '_', 'S', 'Y', 'S', 'C', 'A', 'L', 'L', '_', 'T', 'R', 'A', 'C', 'E', 0}
	var file [_MAX_PATH + 1]uint16
	if stdcall3(_GetEnvironmentVariableW, uintptr(unsafe.Pointer(&envName[0])), uintptr(unsafe.Pointer(&file[0])), uintptr(len(file))) == 0 {
		// Disable syscall tracing.
		syscallTraceFileHandle = _INVALID_HANDLE_VALUE
		return false
	}
	const (
		FILE_APPEND_DATA      = 0x00000004
		FILE_SHARE_READ       = 0x00000001
		FILE_SHARE_WRITE      = 0x00000002
		FILE_SHARE_DELETE     = 0x00000004
		OPEN_ALWAYS           = 4
		FILE_ATTRIBUTE_NORMAL = 0x80
	)
	// Create the file with FILE_APPEND_DATA access to ensure that
	// writes are atomic.
	syscallTraceFileHandle = stdcall7(_CreateFileW, uintptr(unsafe.Pointer(&file[0])), FILE_APPEND_DATA, FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0)
	if syscallTraceFileHandle == _INVALID_HANDLE_VALUE {
		println("runtime: CreateFileA failed; errno=", getlasterror())
		throw("failed to create syscall trace file")
	}
	traceStaticDependencies()
	return true
}

// traceStaticDependencies logs the static dependencies of the runtime.
func traceStaticDependencies() {
	k32 := windowsLoadSystemLib([]uint16{'k', 'e', 'r', 'n', 'e', 'l', '3', '2', 0})
	if k32 == 0 {
		throw("kernel32.dll not found")
	}
	// Some syscalls are dynamically imported by the Windows loader
	// instead of dynamically loaded by the Go runtime. Reading the
	// binary PE important table is too much work, even more considering
	// that we can't allocate memory here. An easier approach is to hardcode
	// the syscalls we know are statically imported. This list can be autogenerated
	// and ensure that is always in sync by running a dedicated test.
	for i := range staticSyscalls {
		syscallTrace(k32, unsafe.StringData(staticSyscalls[i]))
	}
}
